/*
 * Copyright (c) 2018 Xilinx 
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file
 * video Multi Scaler IP (in ABR mode) with Xilinx Media Accelerator
 */

#include <stdio.h>
#include <xma.h>
#include <xmaplugin.h>
#include <xrm.h>

#include "libavutil/attributes.h"
#include "libavutil/internal.h"
#include "libavutil/mem.h"
#include "libavutil/opt.h"
#include "libavutil/imgutils.h"
#include "libavutil/pixfmt.h"

#include "avfilter.h"
#include "formats.h"
#include "internal.h"
#include "video.h"
#include <xvbm.h>
#include <dlfcn.h>

#include "../xmaPropsTOjson.h"

#define MAX_OUTS            8
#define MAX_PARAMS          4
#define VCU_STRIDE_ALIGN    256
#define VCU_HEIGHT_ALIGN    64

#define MAX_INPUT_WIDTH     1920
#define MAX_INPUT_HEIGHT    1080
#define MAX_INPUT_PIXELS    (MAX_INPUT_WIDTH * MAX_INPUT_HEIGHT)

#define ALIGN(width,align) (((width) + (align) - 1) & ~((align) - 1))

#undef DUMP_OUT_FRAMES
#undef DUMP_FRAME_PARAM

#ifdef DUMP_OUT_FRAMES
FILE *outfp = NULL;
FILE *yfp = NULL;
#endif

typedef enum {
    SC_SESSION_ALL_RATE = 0,
    SC_SESSION_FULL_RATE,
    SC_MAX_SESSIONS
} MultiScalerSessionType;

typedef struct MultiScalerContext {
    const AVClass    *class;
    int               nb_outputs;
    int               out_width[MAX_OUTS];
    int               out_height[MAX_OUTS];
    char             *out_format[MAX_OUTS];
    char             *out_rate[MAX_OUTS];
    unsigned int      fps;
    int              *copyOutLink;
    int               flush;
    int               send_status;
    int               frames_out;
    uint32_t          enable_pipeline;

    int               log_level;
    int               latency_logging;
    int               num_sessions;
    int               session_frame;
    uint64_t          p_mixrate_session;
    int               session_nb_outputs[SC_MAX_SESSIONS];
    char              sc_param_name[MAX_PARAMS][50];
    XmaParameter      sc_params[MAX_PARAMS];
    XmaScalerSession *session[SC_MAX_SESSIONS];
    xrmContext       *xrm_ctx;
    xrmCuResource     scalerCuRes[SC_MAX_SESSIONS];
    int               xrm_scalres_count;
    int               xrm_reserve_id;
} MultiScalerContext;


static int multi_scaler_xma_filter_frame(AVFilterLink *link, AVFrame *frame);
static int output_config_props(AVFilterLink *outlink);
static int validate_rate_config(MultiScalerContext *ctx);
static int get_num_scaler_sessions(MultiScalerContext *ctx);
static int get_num_full_rate_outputs(MultiScalerContext *ctx);
static void write_session_log(MultiScalerContext *ctx);

static int validate_rate_config(MultiScalerContext *ctx)
{
    int i, ret;
    int count = 0;

    //All outputs @half-rate not supported
    for (i = 0; i < ctx->nb_outputs; ++i) {
        if (strcmp(ctx->out_rate[i], "half") == 0) {
            count += 1;
        }
    }
    ret = ((ctx->nb_outputs == count) ? -1 : 0);
    return (ret);
}

static int get_num_scaler_sessions(MultiScalerContext *ctx)
{
    int i;
    int count = 1;

    /* default = 1 session - full rate
       However if Mix out_rate is found then 2 sessions will
       be created to allow for frame drops.
    */
    for (i = 0; i < ctx->nb_outputs; ++i) {
        if (strcmp(ctx->out_rate[i], "full") != 0) {
            count = 2;
            break;
        }
    }
    return (count);
}

static int get_num_full_rate_outputs(MultiScalerContext *ctx)
{
    int i;
    int count = 0;

    for (i = 0; i < ctx->nb_outputs; ++i) {
        if (strcmp(ctx->out_rate[i], "full") == 0) {
            count += 1;
        }
    }
    return (count);
}

static void write_session_log(MultiScalerContext *ctx)
{
    int i, count;

    av_log(ctx, AV_LOG_INFO, "  Multi-Scaler Session Configuration\n");
    av_log(ctx, AV_LOG_INFO, "---------------------------------------\n");
    av_log(ctx, AV_LOG_INFO, "Num Sessions = %d\n\n", ctx->num_sessions);

    for (count = 0; count < ctx->num_sessions; ++count) {
        av_log(ctx, AV_LOG_INFO, "Session:  %d\n", count);
        if (ctx->num_sessions > 1) {
            av_log(ctx, AV_LOG_INFO, "Type   :  %s\n", ((count) ? "FULL RATE ONLY" : "HALF RATE"));
        } else {
            av_log(ctx, AV_LOG_INFO, "Type   :  %s\n", "ALL RATE");
        }
        av_log(ctx, AV_LOG_INFO, "Num Out:  %d\n", ctx->session_nb_outputs[count]);
        for (i = 0; i < ctx->session_nb_outputs[count]; ++i) {
            av_log(ctx, AV_LOG_INFO, "out_%d :  (%4d x %4d) @%d fps\n", i, ctx->out_width[i], ctx->out_height[i], ctx->fps);
        }
        av_log(ctx, AV_LOG_INFO, "--------------------------\n");
    }
}
///////////////////////////////////////


static XmaFormatType get_xma_format (enum AVPixelFormat av_format)
{
    switch (av_format) {
        case AV_PIX_FMT_NV12:
        case AV_PIX_FMT_VCU_NV12:
            return XMA_VCU_NV12_FMT_TYPE;
        case AV_PIX_FMT_BGR24:
            return XMA_RGB888_FMT_TYPE;
        default:
            av_log (NULL, AV_LOG_ERROR, "[%s][%d]ERROR : unsupported format %d\n", __func__, __LINE__, av_format);
            return XMA_NONE_FMT_TYPE;
    }
}

static int get_stride (enum AVPixelFormat av_format, int width)
{
    switch (av_format) {
        case AV_PIX_FMT_VCU_NV12:
        case AV_PIX_FMT_NV12:
            return ALIGN(width, VCU_STRIDE_ALIGN);
        case AV_PIX_FMT_BGR24:
            return width * 3;
        default:
            av_log (NULL, AV_LOG_ERROR, "[%s][%d]ERROR : unsupported format %d\n", __func__, __LINE__, av_format);
            return XMA_NONE_FMT_TYPE;
    }
}

static int _allocate_xrm_scaler_cu(AVFilterContext *ctx, XmaScalerProperties *props)
{
    int  ret = -1;
    char pluginName[XRM_MAX_NAME_LEN];
    int scal_load=0, func_id = 0;
    char *err;
    xrmPluginFuncParam param;
    void *handle;
    void (*convertXmaPropsToJson)(void* props, char* funcName, char* jsonJob);
    xrmCuProperty scalerCuProp;

    MultiScalerContext  *s = ctx->priv;

    if (s->xrm_reserve_id > 0) {

        //XRM scaler plugin load calculation
        memset(&param, 0, sizeof(xrmPluginFuncParam));
        handle = dlopen("/opt/xilinx/xrm/plugin/libxmaPropsTOjson.so", RTLD_NOW );
        if (!handle) {
            av_log(ctx, AV_LOG_ERROR, "Unable to load libxmaPropsTOjson.so  - %s\n", dlerror());
            return -1;
        }
        dlerror(); /* clear error code */
        convertXmaPropsToJson = dlsym(handle, "convertXmaPropsToJson");
        if ((err = dlerror()) != NULL) {
             av_log(ctx, AV_LOG_ERROR, "convertXmaPropsToJson symbol not found\n");
             return -1;
        }

        (*convertXmaPropsToJson) (props, (char *)"SCALER", param.input);
        dlclose(handle);
        strcpy(pluginName, "xrmU30ScalPlugin");
        if (xrmExecPluginFunc(s->xrm_ctx, pluginName, func_id, &param) != XRM_SUCCESS)
           av_log(ctx, AV_LOG_ERROR, "scal plugin function %d, fail to run the function\n", func_id);
        else {
           scal_load = atoi((char*)(strtok(param.output, " ")));
        }

        //XRM scaler cu allocation
        memset(&scalerCuProp,                        0, sizeof(xrmCuProperty));
        memset(&s->scalerCuRes[s->xrm_scalres_count], 0, sizeof(xrmCuResource));

        strcpy(scalerCuProp.kernelName,  "scaler");
        strcpy(scalerCuProp.kernelAlias, "SCALER_MPSOC");
        scalerCuProp.devExcl     = false;
        scalerCuProp.requestLoad = scal_load;
        scalerCuProp.poolId      = s->xrm_reserve_id;

        ret = xrmCuAlloc(s->xrm_ctx, &scalerCuProp, &s->scalerCuRes[s->xrm_scalres_count]);
        if (ret != 0) {
            av_log(ctx, AV_LOG_ERROR, "xrm_allocation: fail (err_code=%d) to allocate scaler cu from reserve id %d\n",
                            ret, s->xrm_reserve_id);
            return -1;
        }

        //Set XMA plugin SO and device index
        props->plugin_lib     = s->scalerCuRes[s->xrm_scalres_count].kernelPluginFileName;
        props->dev_index      = s->scalerCuRes[s->xrm_scalres_count].deviceId;
        props->cu_index       = s->scalerCuRes[s->xrm_scalres_count].cuId;
        props->channel_id     = s->scalerCuRes[s->xrm_scalres_count].channelId;
        props->ddr_bank_index = -1;//XMA to select the ddr bank based on xclbin meta data
        if (s->log_level) {
            xrmCuStat scalerCuStat;
            int ret;

            memset(&scalerCuStat, 0, sizeof(xrmCuStat));
            ret = xrmCuCheckStatus(s->xrm_ctx, &s->scalerCuRes[s->xrm_scalres_count], &scalerCuStat);
            printf("Allocated scaler cu: \n");
            printf("   deviceId     = %d\n",  s->scalerCuRes[s->xrm_scalres_count].deviceId);
            printf("   cuId         = %d\n",  s->scalerCuRes[s->xrm_scalres_count].cuId);
            printf("   channelId    = %d\n",  s->scalerCuRes[s->xrm_scalres_count].channelId);
            printf("   cuType       = %d\n",  s->scalerCuRes[s->xrm_scalres_count].cuType);
            printf("   poolId       = %lu\n", s->scalerCuRes[s->xrm_scalres_count].poolId);
            printf("   channelLoad  = %d\n",  s->scalerCuRes[s->xrm_scalres_count].channelLoad);
            if (ret != 0) {
                printf("   Used Load    = xrmCuCheckStatus: fail to check status of scaler cu\n");
            } else {
                printf("   Resvd Load   = %d\n", scalerCuStat.usedLoad);
                printf("   Avail Load   = %d\n", 100-scalerCuStat.usedLoad);
            }
            printf("Note: Channel load is counted against reserved load\n");
        }
    } else {
        //TODO:set default values
        //Set XMA plugin SO and device index
        props->plugin_lib     = (char *)"/opt/xilinx/xma_plugins/libxmultiscaler.so";
        props->dev_index      = 0;
        props->ddr_bank_index = -1;//XMA to select the ddr bank based on xclbin meta data
        props->cu_index       = 0;
        props->channel_id     = 0;
    }

    return 0;
}

static av_cold int multi_scaler_xma_init(AVFilterContext *ctx)
{
    int i = 0;
    MultiScalerContext *s = ctx->priv;
    s->frames_out = 0;

#ifdef DUMP_OUT_FRAMES
    outfp = fopen ("outframes.yuv", "w+");
    yfp = fopen ("outframes_y.yuv", "w+");
#endif

    for (i = 0; i < s->nb_outputs; i++) {
        char name[32];
        AVFilterPad pad = { 0 };

        snprintf(name, sizeof(name), "output%d", i);
        pad.type = ctx->filter->inputs[0].type;
        pad.name = av_strdup(name);
        if (!pad.name)
            return AVERROR(ENOMEM);
        pad.config_props = output_config_props;
        ff_insert_outpad(ctx, i, &pad);
    }
    return 0;
}

static av_cold void multi_scaler_xma_uninit(AVFilterContext *ctx)
{
    int i;
    MultiScalerContext *s = ctx->priv;

    if (getenv("XRM_RESERVE_ID") && s->xrm_ctx) {

       //XRM scaler de-allocation
       for (int idx=0; idx <= s->xrm_scalres_count; idx++) {
             if (!(xrmCuRelease(s->xrm_ctx, &s->scalerCuRes[idx])))
                printf("fail to release scaler HW cu idx=%d\n",idx);
       } 
       
       if (xrmDestroyContext(s->xrm_ctx) != XRM_SUCCESS)
        printf("XRM : destroy context failed\n");
    }

#ifdef DUMP_OUT_FRAMES
    fclose (outfp);
    fclose (yfp);
#endif
    for (i = 0; i < ctx->nb_outputs; i++)
        av_freep(&ctx->output_pads[i].name);

       for (int idx=0; idx < s->num_sessions; idx++) {
          if (s->session[idx])
            xma_scaler_session_destroy(s->session[idx]);
       }
}

int output_config_props(AVFilterLink *outlink)
{
    AVFilterContext *ctx = outlink->src;
    MultiScalerContext *s = ctx->priv;
    const int outlink_idx = FF_OUTLINK_IDX(outlink);
    AVFilterLink *out = outlink->src->outputs[outlink_idx];

    out->w = s->out_width[outlink_idx];
    out->h = s->out_height[outlink_idx];
    outlink->sample_aspect_ratio = (AVRational) {1, 1};

    return 0;
}

static int multi_scaler_xma_config_props(AVFilterLink *outlink)
{
    AVFilterContext *ctx = outlink->dst;
    AVFilterLink *inlink = outlink->dst->inputs[0];
    XmaScalerProperties props;
    MultiScalerContext *s = ctx->priv;
    int n = 0, count=0, ret;
    int chan_id = 0;
    s->fps = 25;
    s->p_mixrate_session = 0;

    props.hwscaler_type = XMA_POLYPHASE_SCALER_TYPE;
    strcpy(props.hwvendor_string, "Xilinx");
    props.num_outputs = s->nb_outputs;

    props.input.format = get_xma_format(inlink->format);
    props.input.width  = inlink->w;
    props.input.height = inlink->h;
    props.input.stride = get_stride(inlink->format, inlink->w);

    //Validate input resolution against MAX supported
    if ((inlink->w > MAX_INPUT_WIDTH) || //for landscape use-case
        (inlink->h > MAX_INPUT_WIDTH) || //for portrait use-case
        ((inlink->w * inlink->h) > MAX_INPUT_PIXELS)) {
        av_log (ctx, AV_LOG_ERROR, "MultiScaler Input %4dx%4d exceeds max supported resolution %4dx%4d (or %4dx%4d portrait mode)\n",
                inlink->w, inlink->h, MAX_INPUT_WIDTH, MAX_INPUT_HEIGHT, MAX_INPUT_HEIGHT, MAX_INPUT_WIDTH);
        return -1;
    }

    if (outlink->time_base.den > 0) {
        int fps = outlink->frame_rate.num/outlink->frame_rate.den;
        av_log(ctx, AV_LOG_INFO, "fps set as %d/%d=%d\n", outlink->frame_rate.num,outlink->frame_rate.den, fps);
        s->fps = fps;
    }

    props.input.framerate.numerator   = s->fps;
    props.input.framerate.denominator = 1;

    //When coeffLoad is set to 2, app expects a FilterCoeff.txt to load coefficients from
    for (n=0; n < MAX_OUTS; n++) {
        if (props.output[n].coeffLoad==2) {
            sprintf(props.input.coeffFile, "FilterCoeff.txt");
            break;
        }
    }

    //run-time parameter configuration
    strcpy(s->sc_param_name[0], "enable_pipeline");    s->sc_params[0].name  = s->sc_param_name[0];    s->sc_params[0].type = XMA_UINT32;    s->sc_params[0].length = sizeof(s->enable_pipeline);    s->sc_params[0].value  = &(s->enable_pipeline);
    strcpy(s->sc_param_name[1], "logLevel");           s->sc_params[1].name  = s->sc_param_name[1];    s->sc_params[1].type = XMA_UINT32;    s->sc_params[1].length = sizeof(s->log_level);          s->sc_params[1].value  = &(s->log_level);
    strcpy(s->sc_param_name[2], "MixRate");            s->sc_params[2].name  = s->sc_param_name[2];    s->sc_params[2].type = XMA_UINT64;    s->sc_params[2].length = sizeof(s->p_mixrate_session);  s->sc_params[2].value  = &(s->p_mixrate_session);
    strcpy(s->sc_param_name[3], "latency_logging");    s->sc_params[3].name  = s->sc_param_name[3];    s->sc_params[3].type = XMA_UINT32;    s->sc_params[3].length = sizeof(s->latency_logging);    s->sc_params[3].value  = &(s->latency_logging);
    props.params           = s->sc_params;
    props.param_cnt        = MAX_PARAMS;

    //validate rate configuration params
    ret = validate_rate_config(s);
    if(ret < 0) {
        av_log (ctx, AV_LOG_ERROR, "Multi Scaler Configuration - All outputs at half-rate not supported\n");
        return -1;
    }

    //determine num sessions to create
    s->num_sessions = get_num_scaler_sessions(s);

    //All-rate session includes all outputs
    s->session_nb_outputs[SC_SESSION_ALL_RATE] = s->nb_outputs;

     if (s->num_sessions > 1) {
        s->session_nb_outputs[SC_SESSION_FULL_RATE] = get_num_full_rate_outputs(s);
        // 2 sessions with half input frame-rate each
        s->fps /= 2;
        props.input.framerate.numerator     = s->fps;
        props.input.framerate.denominator   = 1;
    }
    //log session configuration
    if (s->log_level) {
        write_session_log(s);
    }

    //Get XRM Reservation Id
    if (getenv("XRM_RESERVE_ID")) {
       s->xrm_reserve_id = atoi(getenv("XRM_RESERVE_ID"));

       //create XRM local context
       s->xrm_ctx = (xrmContext *)xrmCreateContext(XRM_API_VERSION_1);
       if (ctx == NULL) {
          av_log (ctx, AV_LOG_ERROR, "create local XRM context failed\n");
          return -1;
       }
    } else {
        s->xrm_reserve_id = -1;
    }

    for (count = 0; count < s->num_sessions; ++count) {
        props.num_outputs = s->session_nb_outputs[count];

        for (chan_id = 0; chan_id < props.num_outputs; chan_id++) {
            props.output[chan_id].format         =  get_xma_format(av_get_pix_fmt(s->out_format[chan_id]));
            props.output[chan_id].bits_per_pixel = av_get_bits_per_pixel(av_pix_fmt_desc_get(av_get_pix_fmt(s->out_format[chan_id])));
            props.output[chan_id].width          = s->out_width[chan_id];
            props.output[chan_id].height         = s->out_height[chan_id];
            props.output[chan_id].stride         = get_stride(av_get_pix_fmt(s->out_format[chan_id]), s->out_width[chan_id]);
            props.output[chan_id].coeffLoad      = 0;
            props.output[chan_id].framerate.numerator   = props.input.framerate.numerator;
            props.output[chan_id].framerate.denominator = props.input.framerate.denominator;
        }

        /*----------------------------------------------------
          Allocate scaler resource from XRM reserved resource
         ----------------------------------------------------*/
        s->xrm_scalres_count = count;
        if(_allocate_xrm_scaler_cu(ctx, &props) < 0) {
            av_log(ctx, AV_LOG_ERROR, "resource allocation failed\n");
            return -1;
        }

        s->session[count] = xma_scaler_session_create(&props);
        if (!s->session[count]) {
            av_log(ctx, AV_LOG_ERROR, "session %d creation failed.\n", count);
            return -1;
        }
        s->p_mixrate_session = (uint64_t)s->session[count]; //send first session handle to next session
    }
    s->session_frame = 0; //start with even frame

    return 0;
}

void xma_multiscaler_filter_flush(AVFilterLink *link)
{
    AVFilterLink *inlink = link->dst->inputs[0];
    AVFilterContext *ctx = link->dst;
    MultiScalerContext *s = ctx->priv;
    int ret = s->send_status;
    int rtt = -1;
    int count = 0;
    int *outLink = (int *)link;
    AVFrame *nframe = av_frame_alloc();

    nframe->format = AV_PIX_FMT_NV12; // sending dummy fixed format
    nframe->width  = inlink->w;
    nframe->height = inlink->h;

    /* creating dummy AVFrame */
    rtt =  av_frame_get_buffer(nframe, VCU_STRIDE_ALIGN);
    if (rtt<0) {
        av_log(ctx, AV_LOG_ERROR, "failed to create dummy AV frame\n");
        return;
    }

    if (outLink == s->copyOutLink) {
        s->flush        = 1;
        nframe->data[0] = NULL;
        nframe->data[1] = NULL;
        nframe->data[2] = NULL;
  
        //flush pipeline for all sessions
        for (count = 0; count < s->num_sessions; ++count) { 
            while (ret != XMA_EOS) {
                multi_scaler_xma_filter_frame(link, nframe);
                ret = s->send_status;
            }
            //reset status for last frame in last session
            ret = XMA_SUCCESS;
        }            
    }

    av_freep(&nframe->data[0]);
    av_freep(&nframe->data[1]);
    av_freep(&nframe->data[2]);
    av_frame_free(&nframe);
}

static int
multi_scaler_xma_filter_frame(AVFilterLink *link, AVFrame *in_frame)
{
    AVFilterContext *ctx = link->dst;
    MultiScalerContext *s = ctx->priv;
    XmaFrame *xframe = NULL;
    int ret = 0;
    int i = 0;
    int plane_id;
    int session_num_out = 0;
    AVFrame *a_frame_list[MAX_OUTS] = {0};
    XmaFrame *x_frame_list[MAX_OUTS] = {0};
    XmaFrameData frame_data = {0, };
    XmaFrameProperties frame_props = {0, };
    XmaScalerSession    *curr_session;
    MultiScalerSessionType  session_type;

    s->copyOutLink = (int*)link;

    if (s->num_sessions > 1) {
        //Odd Frame = SC_SESSION_ALL_RATE, Even Frame = SC_SESSION_FULL_RATE
        session_type = ((s->session_frame & 0x01) ? SC_SESSION_ALL_RATE : SC_SESSION_FULL_RATE);
        s->session_frame = (s->session_frame + 1) % SC_MAX_SESSIONS;
    } else {
        session_type = SC_SESSION_ALL_RATE;
    }
    curr_session    = s->session[session_type];
    session_num_out = s->session_nb_outputs[session_type];


    if (AV_PIX_FMT_VCU_NV12 == in_frame->format) {
        xframe = av_frame_get_xma_frame (in_frame);
        xvbm_buffer_refcnt_inc (xframe->data[0].buffer);
        xframe->pts = in_frame->pts; // Not required if previous elements packs pts
    } else {
        if (FFALIGN(in_frame->width, VCU_STRIDE_ALIGN) == in_frame->linesize[0]) {
            // Clone input frame from an AVFrame to an XmaFrame
            frame_props.format = get_xma_format(in_frame->format);
            frame_props.width = FFALIGN(in_frame->width, VCU_STRIDE_ALIGN);
            frame_props.height = FFALIGN(in_frame->height, VCU_HEIGHT_ALIGN);
            frame_props.bits_per_pixel = 8;

            for (plane_id = 0; plane_id < av_pix_fmt_count_planes (in_frame->format); plane_id++) {
                frame_data.data[plane_id] = in_frame->data[plane_id];
            }

            xframe = xma_frame_from_buffers_clone(&frame_props, &frame_data);
        } else {
            uint8_t *dst;
            uint8_t *src;
            uint32_t dst_stride;
            uint32_t src_stride;
            uint32_t i;
            uint32_t bf_width = FFALIGN(in_frame->width, VCU_STRIDE_ALIGN);
            uint32_t bf_height = FFALIGN(in_frame->height, VCU_HEIGHT_ALIGN);

            frame_props.format = get_xma_format(in_frame->format);
            frame_props.width = bf_width;
            frame_props.height = bf_height;
            frame_props.bits_per_pixel = 8;
        	
            xframe = xma_frame_alloc(&frame_props, false);

            dst = xframe->data[0].buffer;
            src = in_frame->data[0];
            dst_stride = bf_width;
            src_stride = in_frame->linesize[0];
            for (i = 0; i < in_frame->height; i++) {
                memcpy (dst, src, in_frame->width);
                dst += dst_stride;
                src += src_stride;
            }
            dst = xframe->data[1].buffer;
            src = in_frame->data[1];
            src_stride = in_frame->linesize[1];
            for (i = 0; i < in_frame->height / 2; i++) {
                memcpy (dst, src, in_frame->width);
                dst += dst_stride;
                src += src_stride;
            }
        }
        xframe->pts = in_frame->pts;
    }

#ifdef DUMP_FRAME_PARAM
    printf ("MultiScaler Input : w = %d, h = %d, fmt = %d, bps = %d, pts = %lld, data[0] = %p, data[1] = %p, data[2] = %p\n",
        frame_props.width, frame_props.height, frame_props.format, frame_props.bits_per_pixel, in_frame->pts,
        frame_data.data[0], frame_data.data[1], frame_data.data[2]);
#endif

    s->send_status = xma_scaler_session_send_frame(curr_session, xframe);

    /* only receive output frame after XMA_SUCESS or XMA_FLUSH_AGAIN */
    if((s->send_status== XMA_SUCCESS) || (s->send_status == XMA_FLUSH_AGAIN)) {
        int xma_ret = XMA_SUCCESS;
        /* Create output frames */
        for (i = 0; i < session_num_out; i++) {
            XmaFrameProperties fprops;
            XmaFrameData fdata;

            ctx->outputs[i]->format = av_get_pix_fmt(s->out_format[i]);

            if (AV_PIX_FMT_VCU_NV12 == av_get_pix_fmt(s->out_format[i])) {
                a_frame_list[i] = av_frame_alloc();
                if (a_frame_list[i] == NULL) {
                    av_log (ctx, AV_LOG_ERROR, "failed to allocate memory...\n");
                    ret = AVERROR(ENOMEM);
                    goto error;
                }
                a_frame_list[i]->data[0] = NULL;

                fprops.format = get_xma_format(av_get_pix_fmt(s->out_format[i]));
                fprops.width = ctx->outputs[i]->w;
                fprops.height = ctx->outputs[i]->h;
                fprops.bits_per_pixel = av_get_bits_per_pixel(av_pix_fmt_desc_get(av_get_pix_fmt(s->out_format[i])));
                fdata.data[0] = a_frame_list[i]->data[0];
                x_frame_list[i] = xma_frame_from_buffers_clone(&fprops, &fdata);
                x_frame_list[i]->data[0].buffer_type = XMA_DEVICE_BUFFER_TYPE;
            } else {
                a_frame_list[i] = ff_get_video_buffer(ctx->outputs[i], FFALIGN(ctx->outputs[i]->w, VCU_STRIDE_ALIGN), FFALIGN(ctx->outputs[i]->h, VCU_HEIGHT_ALIGN));
                if (a_frame_list[i] == NULL) {
                    av_log (ctx, AV_LOG_ERROR, "failed to allocate output frame...\n");
                    ret = AVERROR(ENOMEM);
                    goto error;
                }
                
                a_frame_list[i]->width = ctx->outputs[i]->w;
                a_frame_list[i]->height = ctx->outputs[i]->h;

                fprops.format = get_xma_format(av_get_pix_fmt(s->out_format[i]));
                fprops.width = FFALIGN(ctx->outputs[i]->w, VCU_STRIDE_ALIGN);
                fprops.height = FFALIGN(ctx->outputs[i]->h, VCU_HEIGHT_ALIGN);
                fprops.bits_per_pixel = av_get_bits_per_pixel(av_pix_fmt_desc_get(av_get_pix_fmt(s->out_format[i])));

                for (plane_id = 0; plane_id < av_pix_fmt_count_planes (av_get_pix_fmt(s->out_format[i])); plane_id++) {
                    fdata.data[plane_id] = a_frame_list[i]->data[plane_id];
                }
                x_frame_list[i] = xma_frame_from_buffers_clone(&fprops, &fdata);
            }
        }

        xma_ret = xma_scaler_session_recv_frame_list(curr_session, x_frame_list);
        if (xma_ret != XMA_SUCCESS) {
            av_log (ctx, AV_LOG_ERROR, "failed to receive frame list from XMA plugin\n");
            ret = AVERROR_UNKNOWN;
            goto error;
        }

        for (i = 0; i < session_num_out; i++) {
            av_frame_copy_props(a_frame_list[i], in_frame);
            a_frame_list[i]->width = ctx->outputs[i]->w;
            a_frame_list[i]->height = ctx->outputs[i]->h;
            a_frame_list[i]->pts = x_frame_list[i]->pts;
            a_frame_list[i]->format  = av_get_pix_fmt(s->out_format[i]);

            if (AV_PIX_FMT_VCU_NV12 == av_get_pix_fmt(s->out_format[i])) {
                a_frame_list[i]->linesize[0] = a_frame_list[i]->linesize[1] = ALIGN(a_frame_list[i]->width, VCU_STRIDE_ALIGN);
                ret = av_frame_clone_xma_frame (a_frame_list[i], x_frame_list[i]);
                if (ret)
                    goto error;
            } else {
                //set the stride
                for (plane_id = 0; plane_id < av_pix_fmt_count_planes (av_get_pix_fmt(s->out_format[i])); plane_id++)
                    a_frame_list[i]->linesize[plane_id] = x_frame_list[i]->frame_props.width;
            }

#ifdef DUMP_OUT_FRAMES
            {
              int written = fwrite (a_frame_list[i]->data[0], 1, (2048*1088*3)>>1, outfp);
              printf ("written %d bytes\n", written);
              //written = fwrite (a_frame_list[i]->data[1], 1, (2048*1080) >> 1, outfp);
              //printf ("written %d bytes\n", written);
            }
#endif
#ifdef DUMP_FRAME_PARAM
            printf ("Output[%d] : w = %d, h = %d, fmt = %d, pts = %lld, linesize[0] = %d,"
                "linesize[1] = %d, linesize[2] = %d, data[0] = %p, data[1]= %p, data[2] = %p\n",
                i, a_frame_list[i]->width, a_frame_list[i]->height, a_frame_list[i]->format, a_frame_list[i]->pts,
                a_frame_list[i]->linesize[0], a_frame_list[i]->linesize[1], a_frame_list[i]->linesize[2],
                a_frame_list[i]->data[0], a_frame_list[i]->data[1], a_frame_list[i]->data[2]);
#endif

            ret = ff_filter_frame(ctx->outputs[i], a_frame_list[i]);
            if (ret < 0) {
                av_log(ctx, AV_LOG_ERROR, "ff_filter_frame failed: ret=%d\n", ret);
                goto error;
            }
            xma_frame_free(x_frame_list[i]);
        }
        s->frames_out++;
    }

    xma_frame_free (xframe);
    if (s->flush == 0)
        av_frame_free(&in_frame);

    return 0;

error:
    if (xframe)
        xma_frame_free (xframe);

    if (s->flush == 0) 
        av_frame_free(&in_frame);

    for (i = 0; i < session_num_out; i++)
        if (x_frame_list[i])
            xma_frame_free(x_frame_list[i]);

    return ret;
}

static int query_formats(AVFilterContext *ctx)
{
    static const enum AVPixelFormat pix_fmts[] = {
        AV_PIX_FMT_VCU_NV12,
        AV_PIX_FMT_NV12, // 25
//        AV_PIX_FMT_BGR24, //3
        AV_PIX_FMT_NONE,
    };
    int res, chan_id;
    MultiScalerContext *s = ctx->priv;
    AVFilterFormats *formats = ff_make_format_list(pix_fmts);

    if (!formats)
        return AVERROR(ENOMEM);
    if (av_get_pix_fmt(s->out_format[0]) == AV_PIX_FMT_NONE)
        return ff_set_common_formats(ctx, formats);
    res = ff_formats_ref(formats, &ctx->inputs[0]->out_formats);
    if (res < 0)
        return res;

    for (chan_id = 0; chan_id < s->nb_outputs; chan_id++) {
        formats = NULL;
        ff_add_format(&formats, av_get_pix_fmt(s->out_format[chan_id]));
        res = ff_formats_ref(formats, &ctx->outputs[chan_id]->in_formats);
        if (res < 0)
            return res;
    }
    return res;
}

#define OFFSET(x) offsetof(MultiScalerContext, x)
#define FLAGS (AV_OPT_FLAG_VIDEO_PARAM | AV_OPT_FLAG_FILTERING_PARAM)
static const AVOption options[] = {
    { "outputs", "set number of outputs", OFFSET(nb_outputs), AV_OPT_TYPE_INT, { .i64 = 8 }, 1, MAX_OUTS, FLAGS },
    { "enable_pipeline", "enable pipelining in multiscaler", OFFSET(enable_pipeline), AV_OPT_TYPE_INT, { }, 0, 1, FLAGS, "enable_pipeline"},
    { "out_1_width", "set width of output 1", OFFSET(out_width[0]), AV_OPT_TYPE_INT, { .i64 = 1600 }, 224, INT_MAX, FLAGS },
    { "out_1_height", "set height of output 1", OFFSET(out_height[0]), AV_OPT_TYPE_INT, { .i64 = 900 }, 126, INT_MAX, FLAGS },
    { "out_1_pix_fmt", "set format of output 1", OFFSET(out_format[0]), AV_OPT_TYPE_STRING, { .str = "nv12"}, CHAR_MIN, CHAR_MAX, FLAGS },
    { "out_1_rate", "set rate of output 1", OFFSET(out_rate[0]), AV_OPT_TYPE_STRING, { .str = "full" }, CHAR_MIN, CHAR_MAX, FLAGS},
    { "out_2_width", "set width of output 2", OFFSET(out_width[1]), AV_OPT_TYPE_INT, { .i64 = 1280 }, 224, INT_MAX, FLAGS },
    { "out_2_height", "set height of output 2", OFFSET(out_height[1]), AV_OPT_TYPE_INT, { .i64 = 720 }, 126, INT_MAX, FLAGS },
    { "out_2_pix_fmt", "set format of output 2", OFFSET(out_format[1]), AV_OPT_TYPE_STRING, { .str = "nv12"}, CHAR_MIN, CHAR_MAX, FLAGS },
    { "out_2_rate", "set rate of output 2", OFFSET(out_rate[1]), AV_OPT_TYPE_STRING, { .str = "full" }, CHAR_MIN, CHAR_MAX, FLAGS},
    { "out_3_width", "set width of output 3", OFFSET(out_width[2]), AV_OPT_TYPE_INT, { .i64 = 800 }, 224, INT_MAX, FLAGS },
    { "out_3_height", "set height of output 3", OFFSET(out_height[2]), AV_OPT_TYPE_INT, { .i64 = 600 }, 126, INT_MAX, FLAGS },
    { "out_3_pix_fmt", "set format of output 3", OFFSET(out_format[2]), AV_OPT_TYPE_STRING, { .str = "nv12"}, CHAR_MIN, CHAR_MAX, FLAGS },
    { "out_3_rate", "set rate of output 3", OFFSET(out_rate[2]), AV_OPT_TYPE_STRING, { .str = "full" }, CHAR_MIN, CHAR_MAX, FLAGS},
    { "out_4_width", "set width of output 4", OFFSET(out_width[3]), AV_OPT_TYPE_INT, { .i64 = 832 }, 224, INT_MAX, FLAGS },
    { "out_4_height", "set height of output 4", OFFSET(out_height[3]), AV_OPT_TYPE_INT, { .i64 = 480 }, 126, INT_MAX, FLAGS },
    { "out_4_pix_fmt", "set format of output 4", OFFSET(out_format[3]), AV_OPT_TYPE_STRING, { .str = "nv12"}, CHAR_MIN, CHAR_MAX, FLAGS },
    { "out_4_rate", "set rate of output 4", OFFSET(out_rate[3]), AV_OPT_TYPE_STRING, { .str = "full" }, CHAR_MIN, CHAR_MAX, FLAGS},
    { "out_5_width", "set width of output 5", OFFSET(out_width[4]), AV_OPT_TYPE_INT, { .i64 = 640 }, 224, INT_MAX, FLAGS },
    { "out_5_height", "set height of output 5", OFFSET(out_height[4]), AV_OPT_TYPE_INT, { .i64 = 480 }, 126, INT_MAX, FLAGS },
    { "out_5_pix_fmt", "set format of output 5", OFFSET(out_format[4]), AV_OPT_TYPE_STRING, { .str = "nv12"}, CHAR_MIN, CHAR_MAX, FLAGS },
    { "out_5_rate", "set rate of output 5", OFFSET(out_rate[4]), AV_OPT_TYPE_STRING, { .str = "full" }, CHAR_MIN, CHAR_MAX, FLAGS},
    { "out_6_width", "set width of output 6", OFFSET(out_width[5]), AV_OPT_TYPE_INT, { .i64 = 480 }, 224, INT_MAX, FLAGS },
    { "out_6_height", "set height of output 6", OFFSET(out_height[5]), AV_OPT_TYPE_INT, { .i64 = 320 }, 126, INT_MAX, FLAGS },
    { "out_6_pix_fmt", "set format of output 6", OFFSET(out_format[5]), AV_OPT_TYPE_STRING, { .str = "nv12"}, CHAR_MIN, CHAR_MAX, FLAGS },
    { "out_6_rate", "set rate of output 6", OFFSET(out_rate[5]), AV_OPT_TYPE_STRING, { .str = "full" }, CHAR_MIN, CHAR_MAX, FLAGS},
    { "out_7_width", "set width of output 7", OFFSET(out_width[6]), AV_OPT_TYPE_INT, { .i64 = 320 }, 224, INT_MAX, FLAGS },
    { "out_7_height", "set height of output 7", OFFSET(out_height[6]), AV_OPT_TYPE_INT, { .i64 = 240 }, 126, INT_MAX, FLAGS },
    { "out_7_pix_fmt", "set format of output 7", OFFSET(out_format[6]), AV_OPT_TYPE_STRING, { .str = "nv12"}, CHAR_MIN, CHAR_MAX, FLAGS },
    { "out_7_rate", "set rate of output 7", OFFSET(out_rate[6]), AV_OPT_TYPE_STRING, { .str = "full" }, CHAR_MIN, CHAR_MAX, FLAGS},
    { "out_8_width", "set width of output 8", OFFSET(out_width[7]), AV_OPT_TYPE_INT, { .i64 = 224 }, 224, INT_MAX, FLAGS },
    { "out_8_height", "set height of output 8", OFFSET(out_height[7]), AV_OPT_TYPE_INT, { .i64 = 224 }, 126, INT_MAX, FLAGS },
    { "out_8_pix_fmt", "set format of output 8", OFFSET(out_format[7]), AV_OPT_TYPE_STRING, { .str = "nv12"}, CHAR_MIN, CHAR_MAX, FLAGS },
    { "out_8_rate", "set rate of output 8", OFFSET(out_rate[7]), AV_OPT_TYPE_STRING, { .str = "full" }, CHAR_MIN, CHAR_MAX, FLAGS},
    { "logLevel", "level for debug message", OFFSET(log_level),  AV_OPT_TYPE_INT, { .i64 = 0 }, 0, 2, FLAGS},
	{ "latency_logging", "Log latency information to syslog", OFFSET(latency_logging), AV_OPT_TYPE_INT, { .i64 = 0 }, 0, 1, FLAGS, "latency_logging" },
    { NULL }
};

#define multi_scaler_xma_options options
AVFILTER_DEFINE_CLASS(multi_scaler_xma);

static const AVFilterPad avfilter_vf_multi_scaler_xma_inputs[] = {
    {
        .name = "default",
        .type = AVMEDIA_TYPE_VIDEO,
        .filter_frame = multi_scaler_xma_filter_frame,
        .config_props = multi_scaler_xma_config_props,
    },
    { NULL }
};

AVFilter ff_vf_multiscale_xma = {
    .name = "multiscale_xma",
    .description = NULL_IF_CONFIG_SMALL("Xilinx Multi Scaler (in ABR mode) using XMA APIs"),
    .priv_size = sizeof(MultiScalerContext),
    .priv_class = &multi_scaler_xma_class,
    .query_formats = query_formats,
    .init = multi_scaler_xma_init,
    .uninit = multi_scaler_xma_uninit,
    .inputs = avfilter_vf_multi_scaler_xma_inputs,
    .outputs = NULL,
    .flags = AVFILTER_FLAG_DYNAMIC_OUTPUTS,
};

